import flywheel
import pandas as pd
import sys
import argparse
from tqdm import tqdm
import re
from pandas.io.json.normalize import nested_to_record
from pandas.api.types import is_list_like
import warnings
from flywheel_bids_tools.utils import unlist_item

UNCLASSIFIED = 0
NO_DATA = 0
VERBOSE = True


def query_fw(project, client):
    """Query the flywheel client for a project name

    This function uses the flywheel API to find the first match of a project
    name. The name must be exact so make sure to type it as is on the
    website/GUI.

    Parameters
    ---------
    project
        The name of the project to search for.
    client
        The flywheel Client class object.

    Returns
    ---------
    project_object
        A flywheel project container object.
    """
    project_object = client.projects.find_first('label={0}'.format(project))

    if project_object is None:
        print("Available projects are:\n")
        for p in client.projects():
            print('%s' % (p.label))
        raise ValueError("Could not find \"{0}\" project on Flywheel!".format(project))
    return(project_object)


def extract_bids_data(acquisitionID, client):
    """Extract the BIDS data of an acquisition

    A helper function to dig into the file.info container
    (a dictionary of dictionaries) and extract the BIDS validity fields.

    Parameters
    -------
    acquisitionID
        The mongoDB hash key to identify the object.
    client
        The flywheel Client class object.

    Returns
    --------
    df
        A table of the bids fields and values.
    """
    # create the acquisition object and pull the niftis
    try:
        acq = client.get(acquisitionID)
    except:
        global NO_DATA
        NO_DATA += 1
        return None
    niftis = [x for x in acq.files if x['type'] == 'nifti']
    # if there are no niftis, return
    if (len(niftis) < 1):
        global UNCLASSIFIED
        UNCLASSIFIED += 1
        return None
    else:
        df = []
        # for each nifti, if the info has a BIDS dict:
        for nii in niftis:
            info = nii['info']
            if 'BIDS' in info.keys() and isinstance(info['BIDS'], dict):
                # also add the acquisition id to the dict for joining purposes
                nii['info']['BIDS']['acquisition.id'] = str(acquisitionID)
                # pull out the bids info
                bids = nii['info']['BIDS']
                # include TR and Series name
                bids.update({'RepetitionTime': nii['info']['RepetitionTime']})
                bids.update({'SeriesDescription': nii['info']['SeriesDescription']})
                # include the classification
                if 'classification' in nii.keys():
                    bids.update(nii.classification)
                df.append(bids)

            else:
                nii['info']['BIDS'] = {"acquisition.id": acquisitionID}
                # pull out the bids info
                bids = nii['info']['BIDS']
                # include the classification
                if 'classification' in nii.keys():
                    bids.update(nii.classification)
                df.append(bids)

        return(df)


def process_acquisition(acq_id, client):
    '''
    Extract an acquisition

    This function extracts an acquisition object and collects the important
    file classification information. These data are processed and returned as
    a pandas dataframe that can then be manipulated

    '''

    # get the acquisition object
    acq = client.get(acq_id)
    if acq is None:
        global UNCLASSIFIED
        UNCLASSIFIED += 1
    # convert to dictionary, and flatten the dictionary to avoid nested dicts
    files = [x.to_dict() for x in acq.files]
    flat_files = [nested_to_record(my_dict, sep='_') for my_dict in files]

    # define desirable columns in regex
    cols = r'(classification)|(^type$)|(^modality$)|(BIDS)|(EchoTime)|(RepetitionTime)|(SequenceName)|(SeriesDescription)'

    # filter the dict keys for the columns names
    flat_files = [
        {k: v for k, v in my_dict.items() if re.search(cols, k)}
        for my_dict in flat_files
        ]

    # add acquisition ID for reference
    for x in flat_files:
        x.update({'acquisition.id': acq_id})

    # to data frame
    df = pd.DataFrame(flat_files)

    # lastly, only pull niftis and dicoms; also convert list to string
    if 'type' in df.columns:
        df = df[df.type.str.contains(r'nifti|dicom', na=False)].reset_index(drop=True)
    if 'BIDS' not in df.columns:
        global NO_DATA
        NO_DATA += 1

    return df


def query_bids_validity(project, client, VERBOSE=True):
    """Query Flywheel for BIDS data

    Main wrapper function for querying, processing, and extracting BIDS
    information for a project

    Parameters
    --------
    project
        Project to query for
    client
        A flywheel project container object
    verbose
        Print progress messages

    Returns
    --------
    merged_data
        A dataframe of the result of the query and processing
    """

    # first, log in to flywheel
    if VERBOSE:
        print("Connecting to flywheel server...")
    assert client
    # query flywheel for the argument
    if VERBOSE:
        print("Querying server...")

    result = query_fw(project, client)

    # get the subjects for this project
    view = client.View(columns='subject')
    subject_df = client.read_view_dataframe(view, result.id)

    # loop through the subjects to extract acquisitions
    if VERBOSE:
        print("Processing acquisitions...")
    sessions = []
    view = client.View(columns='acquisition')
    for ind, row in tqdm(subject_df.iterrows(), total=subject_df.shape[0]):
        session = client.read_view_dataframe(view, row["subject.id"])
        if(session.shape[0] > 0):
            sessions.append(session)

    acquisitions = pd.concat(sessions)

    # loop through the acquisitions to extract the bids validity data
    # note: speed bottleneck here
    if VERBOSE:
        print("Extracting BIDS and MR Classifier information...")
    bids_classifications = []
    for index, row in tqdm(acquisitions.iterrows(), total=acquisitions.shape[0]):

        temp = process_acquisition(row["acquisition.id"], client)
        bids_classifications.append(temp)

    bids_classifications = pd.concat(bids_classifications, sort=False)

    # finally, join the bids classification with the acquisitions
    if VERBOSE:
        print("Tidying and returning the results...")
    merged_data = pd.merge(acquisitions, bids_classifications, how='outer')
    merged_data = merged_data.drop(columns=['acquisition.timestamp', 'acquisition.timezone', 'project.id', 'session.id', 'subject.id'])
    if VERBOSE:
        print("{} acquisitions could not be processed.".format(UNCLASSIFIED))
        print("{} acquisitions do not have BIDS information yet.".format(NO_DATA))
    return(merged_data)


def main():

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        fw = flywheel.Client()
        assert fw, "Your Flywheel CLI credentials aren't set!"

    parser = argparse.ArgumentParser(description=("Use this tool to query Flywheel for a project and write out the acquisitions to a table"))
    parser.add_argument(
        "-proj", "--project",
        help="The project in flywheel to search for",
        nargs="+",
        required=True,
        dest="project"
    )
    parser.add_argument(
        "-output", "--output-file",
        help="The path and name of the output CSV of the query",
        required=True,
        dest="output"
    )
    parser.add_argument(
        "-v", "--verbose",
        help="Print out progress messages and information",
        default=True
    )

    args = parser.parse_args()

    global VERBOSE
    VERBOSE = args.verbose
    project = ' '.join(args.project)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        query_result = query_bids_validity(project, fw)
        query_result = query_result.sort_values(by=["acquisition.id", "acquisition.label"])
        infer_type = lambda x: pd.api.types.infer_dtype(x, skipna=True)
        #list_cols = query_result.apply(infer_type, axis=0) == 'mixed'
        #query_result.loc[:, list_cols] = query_result.loc[:, list_cols].applymap(utils.unlist_item)
        drop_downs = ['classification_Measurement', 'classification_Intent',
            'classification_Features']
        query_result.loc[:, drop_downs] = query_result.loc[:, drop_downs].applymap(unlist_item)
        query_result.to_csv(args.output, index=False)
    print("Done!")


if __name__ == '__main__':
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="flywheel_bids_tools",
    version="0.0.8",
    author="Tinashe M. Tapera",
    author_email="tinashemtapera@gmail.com",
    description="Tools for extracting, editing, and uploading BIDS data from and to Flywheel",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PennBBL/bids-on-flywheel",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points = {
        'console_scripts': [
            'query-bids=flywheel_bids_tools.query_bids:main',
            'upload-bids=flywheel_bids_tools.upload_bids:main',
            'group-query=flywheel_bids_tools.group_query:main',
            'ungroup-query=flywheel_bids_tools.ungroup_query:main',
            'query-gears=flywheel_bids_tools.query_gears:main',
            'run-gear=flywheel_bids_tools.run_gear:main'
        ],
    }
)



import sys
import os
import pandas as pd
import numpy as np
import re
import flywheel
import math
import numbers
import datetime
import argparse
from flywheel_bids_tools.utils import relist_item, get_unequal_cells
from tqdm import tqdm


ERROR_MESSAGES = []


def read_flywheel_csv(fpath, required_cols=['acquisition.id']):
    '''
    Read in a CSV and also ensure it's one of ours

    Input:
        fpath: path to the file
        required_cols: list of columns to ensure csv is a flywheel query
    Output:
        df: a pandas dataframe
    '''

    df = pd.read_csv(fpath, dtype={'valid':object})

    if not all(elem in df.columns.tolist() for elem in required_cols):
        raise Exception(("It doesn't look like this csv is correctly formatted",
        " for this flywheel editing process!"))

    df = df.sort_values(by=["acquisition.id", "acquisition.label"])
    drop_downs = ['classification_Measurement', 'classification_Intent']
    df.loc[:, drop_downs] = df.loc[:, drop_downs].applymap(relist_item)
    return(df)


def change_checker(user_input, column):
    '''
    When a user has changed a value in a column, check that the input they
    supplied is in the allowed list of acceptable changes. Note this function
    adjusts cases.

    Input:
        user_input: what the user tried to change
        column: the column they tried to change
    Output:
        boolean: True or False whether the change is acceptable on Flywheel
    '''

    drop_down_bool = {
        'ignore': ['true', 'false'],
        'valid': ['true', 'false']
        }

    drop_down_single_fields = {
        'modality': ['', 'mr', 'ct', 'pet', 'us', 'eeg', 'ieeg', 'x-ray',
            'ecg', 'meg', 'nirs']
        }

    drop_down_multi_fields = {
        'classification_measurement': ['ASL', 'B0', 'B1', 'Diffusion',
            'Fingerprinting', 'MT', 'PD', 'Perfusion', 'Spectroscopy',
            'Susceptibility', 'T1', 'T2', 'T2*', 'Velocity'],

        'classification_intent': ['Calibration', 'Fieldmap', 'Functional',
            'Localizer', 'Non-Image', 'Screenshot', 'Shim', 'Structural'],

        'classification_Features': ['3D', 'Compressed-Sensing', 'Derived',
            'Eddy-Current-Corrected', 'Fieldmap-Corrected', 'Gradient-Unwarped',
            'In-Plane', 'Magnitude', 'Motion-Corrected', 'Multi-Band',
            'Multi-Echo', 'Multi-Flip', 'Multi-Shell', 'Phase',
            'Physio-Corrected', 'Quantitative', 'Steady-State']
    }

    string_fields = ['acquisition.label', 'project.label', 'error_message',
        'subject.label', 'folder', 'template',
        'intendedfor', 'mod', 'path', 'rec', 'task', 'run',
        'info_bids_error_message', 'info_sequencename',
        'info_bids_filename', 'type', 'info_bids_folder', 'info_bids_modality',
        'info_bids_path', 'info_bids_template', 'info_seriesdescription',
        'classification_custom']

    numeric_fields = ['info_echotime', 'info_repetitiontime']

    # try boolean drop down option
    if column.lower() in drop_down_bool.keys():
        if str(user_input).lower() == 'true' or str(user_input).lower() == 'false':
            return True
        else:
            ERROR_MESSAGES.append("This field accepts booleans, these can only be written as \"True\" or \"False\"!")
            return False

    # try drop down string option
    elif column.lower() in drop_down_single_fields.keys():
        for field, options in drop_down_single_fields.items():
            if str(user_input).lower() in options:
                return True
            else:
                continue
        ERROR_MESSAGES.append("This field must match one of the available options in the drop-down menu on the website!")
        return False

    # try drop down multi option
    elif column.lower() in drop_down_multi_fields.keys():

        if not isinstance(user_input, list):
            user_input = [user_input]
        if set(user_input) <= set(drop_down_multi_fields[column.lower()]):
            return True
        else:
            ERROR_MESSAGES.append("This field must match one of the available options in the drop-down menu on the website!")
            return False

    # try generic string
    elif column.lower() in string_fields:
        if type(user_input) is str or math.isnan(user_input):
            return True
        else:
            ERROR_MESSAGES.append("This field only accepts strings!")
            return False

    # try generic number
    elif column.lower() in numeric_fields:
        if isinstance(user_input, numbers.Number):
            return True
        else:
            ERROR_MESSAGES.append("This field only accepts numeric types!")
            return False

    # try the filename separately
    elif column.lower() == 'filename':
        # three different types of acceptable file names
        anat1 = re.compile(
            r'^sub-(?P<subject_id>[a-zA-Z0-9]+)(_ses-(?P<session_id>[a-zA-Z0-9]+))?(_acq-(?P<acquisition_label>[a-zA-Z0-9]+))?(_ce-(?P<contrastenhanced_id>[a-zA-Z0-9]+))?(_rec-(?P<reconstruction_id>[a-zA-Z0-9]+))?(_run-(?P<run_id>[a-zA-Z0-9]+))?(_(?P<modality>[a-zA-Z0-9]+))?((?P<suffix>\.nii(\.gz)?))$'
            )

        anat2 = re.compile(
            r'^sub-(?P<subject_id>[a-zA-Z0-9]+)(_ses-(?P<session_id>[a-zA-Z0-9]+))?(_acq-(?P<acquisition_label>[a-zA-Z0-9]+))?(_ce-(?P<contrastenhanced_id>[a-zA-Z0-9]+))?(_rec-(?P<reconstruction_id>[a-zA-Z0-9]+))?(_run-(?P<run_id>[a-zA-Z0-9]+))?(_mod-(?P<modality>[a-zA-Z0-9]+))?(_(?P<suffix>[a-zA-Z0-9]+\.nii(\.gz)?))$'
            )

        func1 = re.compile(
            r'^sub-(?P<subject_id>[a-zA-Z0-9]+)(_ses-(?P<session_id>[a-zA-Z0-9]+))?(_task-(?P<task_label>[a-zA-Z0-9]+))?(_acq-(?P<acquisition_label>[a-zA-Z0-9]+))?(_ce-(?P<contrastenhanced_id>[a-zA-Z0-9]+))?(_dir-(?P<direction>[a-zA-Z0-9]+))?(_rec-(?P<reconstruction_id>[a-zA-Z0-9]+))?(_run-(?P<run_id>[a-zA-Z0-9]+))?(_echo-(?P<echo_id>[a-zA-Z0-9]+))?(_(?P<contrast_label>[a-zA-Z0-9]+))?((?P<suffix>\.nii(\.gz)?))$'
        )

        if bool(anat1.match(str(user_input))):
            return True
        elif bool(anat2.match(str(user_input))):
            return True
        elif bool(func1.match(str(user_input))):
            return True
        else:
            ERROR_MESSAGES.append("This field MUST be a BIDS compliant name!")
            return False

    # can't edit acquisition ID tho!
    elif column == 'acquisition.id':
        ERROR_MESSAGES.append("You cannot edit the acquisition ID!")
        return False
    else:
        raise Exception("Column {0} not recognised!".format(column))


def validate_on_unequal_cells(indices_list, changed_df):
    '''
    Loop over list of row-column pairs and run the change checker on each

    Input:
        indices_list: array of the row indices and column indices
        changed_df: the df to run the change checker on
    Output:
        valid: boolean on whether the df passed the change checker
    '''

    valid = []
    for pair in indices_list:
        valid.append(change_checker(
            changed_df.iloc[pair[0], pair[1]],
            changed_df.columns[pair[1]]
            ))

    if all(valid):
        return True
    else:
        print("The following changes don't seem to be valid for this data:")
        for x in range(len(valid)):
            if valid[x] is False:
                print("")
                print("Row {}, Column {}, \"{}\"".format(
                    indices_list[x][0]+1,
                    indices_list[x][1]+1,
                    changed_df.iloc[indices_list[x][0], indices_list[x][1]]
                    ))
                print(ERROR_MESSAGES.pop(0))
        return False


def upload_to_flywheel(modified_df, change_index, client):
    '''
    If the changes are valid, upload them to flywheel
    '''

    # loop through each of the row_col indexes of changes
    for pair in tqdm(change_index, total=len(change_index)):

        # get the acquisition id
        acquisition = modified_df.loc[pair[0], 'acquisition.id']
        file_type = modified_df.loc[pair[0], 'type']

        # get the flywheel object of the acquisition
        fw_object = client.get(str(acquisition))

        # create the update dictionary
        column_list = modified_df.columns[pair[1]].split("_")
        value = modified_df.iloc[pair[0], pair[1]]
        if value is float('nan'):
            continue
        update = create_nested_fw_dict(column_list, value)

        f = [f for f in fw_object.files if f.type == file_type][0]
        if 'info' in update.keys():
            f.update_info(update['info'])
        if 'classification' in update.keys():
            f.replace_classification(update['classification'])

    return


def create_nested_fw_dict(tree_list, value):

    if tree_list:
        return {tree_list[0]: create_nested_fw_dict(tree_list[1:], value)}
    return value


def main():

    fw = flywheel.Client()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-orig",
        help="Path to the original flywheel query CSV",
        dest="original",
        required=True
    )
    parser.add_argument(
        "-mod",
        help="Path to the modified flywheel query CSV",
        dest="modified",
        required=True
    )

    args = parser.parse_args()

    # original df
    df_original = read_flywheel_csv(args.original)
    # edited df
    df_modified = read_flywheel_csv(args.modified)

    # check for equality of each cell between the original and modified
    unequal = get_unequal_cells(df_original, df_modified)
    # if any unequal, assess the validity of the modification
    res = validate_on_unequal_cells(unequal, df_modified)

    if len(ERROR_MESSAGES) is 0 and res is True:
        print("Changes appear to be valid! Uploading...")
        upload_to_flywheel(df_modified, unequal, fw)
        print("Done!")
        sys.exit(0)
    else:
        print("Exiting...")
        sys.exit(0)


if __name__ == '__main__':
    main()
